<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\Controller;

// Prevent direct access
defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\ControllerEvents;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\GetRedirectionAware;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\RegisterControllerTasks;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\ReturnURLAware;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\ReusableModels;
use Akeeba\Component\LoginGuard\Administrator\Event\NotifyActionLog;
use Akeeba\Component\LoginGuard\Administrator\Event\SaveSetup;
use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Model\BackupcodesModel;
use Akeeba\Component\LoginGuard\Administrator\Model\MethodModel;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Controller\BaseController as BaseControllerAlias;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\MVC\Model\BaseModel;
use Joomla\CMS\Router\Route;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Input\Input;
use RuntimeException;

class MethodController extends BaseControllerAlias
{
	use ControllerEvents;
	use GetRedirectionAware;
	use RegisterControllerTasks;
	use ReturnURLAware;
	use ReusableModels;

	/**
	 * The default model for this MVC triad. Something that Joomla fails to take into account...
	 *
	 * @var   BaseModel
	 */
	protected $default_model = null;

	public function __construct($config = [], MVCFactoryInterface $factory = null, ?CMSApplication $app = null, ?Input $input = null)
	{
		// We have to tell Joomla what is the name of the view, otherwise it defaults to the name of the *component*.
		$config['default_view'] = 'method';
		$config['default_task'] = 'add';

		parent::__construct($config, $factory, $app, $input);
	}

	/**
	 * Execute a task by triggering a Method in the derived class.
	 *
	 * @param   string  $task    The task to perform. If no matching task is found, the '__default' task is executed, if
	 *                           defined.
	 *
	 * @return  mixed   The value returned by the called Method.
	 *
	 * @throws  Exception
	 * @since   3.0
	 */
	public function execute($task)
	{
		// I wish I could use a single controller for plural and singular views - but I can't.
		if ($task == 'display')
		{
			$task = 'add';
		}

		return parent::execute($task);
	}

	/**
	 * Add a new MFA Method
	 *
	 * @param   bool   $cachable     Can this view be cached
	 * @param   array  $urlparams    An array of safe url parameters and their variable types, for valid values see
	 *                               {@link JFilterInput::clean()}.
	 *
	 * @return  self
	 */
	public function add($cachable = false, $urlparams = []): self
	{
		$this->assertLoggedInUser();

		// Make sure I am allowed to edit the specified user
		$user_id = $this->input->getInt('user_id', null);
		$user    = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($user_id);

		$this->_assertCanEdit($user);

		// Also make sure the Method really does exist
		$method = $this->input->getCmd('method');
		$this->_assertMethodExists($method);

		/** @var MethodModel $model */
		$model = $this->getModel('Method', 'Administrator');
		$model->setState('method', $method);

		// Pass the return URL to the view
		$returnURL  = $this->input->getBase64('returnurl');
		$viewLayout = $this->input->get('layout', 'default', 'string');
		$view       = $this->getView();
		$view->setLayout($viewLayout);
		$view->returnURL = $returnURL;
		$view->user      = $user;
		$view->document  = $this->app->getDocument();

		$view->setModel($model, true);

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginguardControllerMethodBeforeAdd', [$user, $method])
		);

		$view->display();

		return $this;
	}

	/**
	 * Edit an existing MFA Method
	 *
	 * @param   bool   $cachable     Can this view be cached
	 * @param   array  $urlparams    An array of safe url parameters and their variable types, for valid values see
	 *                               {@link JFilterInput::clean()}.
	 *
	 * @return  self   The current JControllerLegacy object to support chaining.
	 */
	public function edit($cachable = false, $urlparams = []): self
	{
		$this->assertLoggedInUser();

		// Make sure I am allowed to edit the specified user
		$user_id = $this->input->getInt('user_id', null);
		$user    = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($user_id);
		$this->_assertCanEdit($user);

		// Also make sure the Method really does exist
		$id     = $this->input->getInt('id');
		$record = $this->_assertValidRecordId($id, $user);

		if ($id <= 0)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		/** @var MethodModel $model */
		$model = $this->getModel('Method', 'Administrator');
		$model->setState('id', $id);

		// Pass the return URL to the view
		$returnURL  = $this->input->getBase64('returnurl');
		$viewLayout = $this->input->get('layout', 'default', 'string');
		$view       = $this->getView();
		$view->setLayout($viewLayout);
		$view->returnURL = $returnURL;
		$view->user      = $user;
		$view->document  = $this->app->getDocument();

		$view->setModel($model, true);

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginguardControllerMethodBeforeEdit', [$id, $user])
		);

		$view->display();

		return $this;
	}

	/**
	 * Regenerate backup codes
	 *
	 * @param   bool   $cachable     Can this view be cached
	 * @param   array  $urlparams    An array of safe url parameters and their variable types, for valid values see
	 *                               {@link JFilterInput::clean()}.
	 *
	 * @return  self   The current JControllerLegacy object to support chaining.
	 */
	public function regenbackupcodes($cachable = false, $urlparams = []): self
	{
		$this->assertLoggedInUser();

		$this->checkToken($this->input->getMethod());

		// Make sure I am allowed to edit the specified user
		$user_id = $this->input->getInt('user_id', null);
		$user    = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($user_id);
		$this->_assertCanEdit($user);

		/** @var BackupcodesModel $model */
		$model = $this->getModel('Backupcodes', 'Administrator');
		$model->regenerateBackupCodes($user);

		$backupCodesRecord = $model->getBackupCodesRecord($user);

		// Redirect
		$redirectUrl = 'index.php?option=com_loginguard&task=method.edit&user_id=' . $user_id . '&id=' . $backupCodesRecord->id;
		$returnURL   = $this->input->getBase64('returnurl');

		if (!empty($returnURL))
		{
			$redirectUrl .= '&returnurl=' . $returnURL;
		}

		$this->setRedirect(Route::_($redirectUrl, false));

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginguardControllerMethodAfterRegenbackupcodes')
		);

		return $this;
	}

	/**
	 * Delete an existing MFA Method
	 *
	 * @param   bool   $cachable     Can this view be cached
	 * @param   array  $urlparams    An array of safe url parameters and their variable types, for valid values see
	 *                               {@link JFilterInput::clean()}.
	 *
	 * @return  self   The current JControllerLegacy object to support chaining.
	 */
	public function delete($cachable = false, $urlparams = [])
	{
		$this->assertLoggedInUser();

		$this->checkToken($this->input->getMethod());

		// Make sure I am allowed to edit the specified user
		$user_id = $this->input->getInt('user_id', null);
		$user    = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($user_id);
		$this->_assertCanEdit($user);

		// Also make sure the Method really does exist
		$id     = $this->input->getInt('id');
		$record = $this->_assertValidRecordId($id, $user);

		if ($id <= 0)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		$type    = null;
		$message = null;

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginguardControllerMethodBeforeDelete', [$id, $user])
		);

		try
		{
			$record->delete();
		}
		catch (Exception $e)
		{
			$message = $e->getMessage();
			$type    = 'error';
		}

		// Redirect
		$url       = Route::_('index.php?option=com_loginguard&task=methods.display&user_id=' . $user_id, false);
		$returnURL = $this->input->getBase64('returnurl');

		if (!empty($returnURL))
		{
			$url = base64_decode($returnURL);
		}

		$this->setRedirect($url, $message, $type);

		return $this;
	}

	/**
	 * Save the MFA Method
	 *
	 * @param   bool   $cachable     Can this view be cached
	 * @param   array  $urlparams    An array of safe url parameters and their variable types, for valid values see
	 *                               {@link JFilterInput::clean()}.
	 *
	 * @return  self   The current JControllerLegacy object to support chaining.
	 */
	public function save($cachable = false, $urlparams = [])
	{
		$this->assertLoggedInUser();

		$this->checkToken($this->input->getMethod());

		// Make sure I am allowed to edit the specified user
		$user_id = $this->input->getInt('user_id', null);
		$user    = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($user_id);
		$this->_assertCanEdit($user);

		// Redirect
		$url       = Route::_('index.php?option=com_loginguard&task=methods.display&user_id=' . $user_id, false);
		$returnURL = $this->input->getBase64('returnurl');

		if (!empty($returnURL))
		{
			$url = base64_decode($returnURL);
		}

		// The record must either be new (ID zero) or exist
		$id     = $this->input->getInt('id', 0);
		$record = $this->_assertValidRecordId($id, $user);

		// If it's a new record we need to read the Method from the request and update the (not yet created) record.
		if ($record->id == 0)
		{
			$methodName = $this->input->getCmd('method');
			$this->_assertMethodExists($methodName);
			$record->method = $methodName;
		}

		/** @var MethodModel $model */
		$model = $this->getModel('Method', 'Administrator');

		// Ask the plugin to validate the input by calling onLoginGuardMfaSaveSetup
		$result = [];
		$input  = $this->app->input;

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginguardControllerMethodBeforeSave', [$id, $user])
		);

		try
		{
			$pluginResults = MfaHelper::triggerEvent(new SaveSetup($record, $input));

			foreach ($pluginResults as $pluginResult)
			{
				$result = array_merge($result, $pluginResult);
			}
		}
		catch (RuntimeException $e)
		{
			// Go back to the edit page
			$nonSefUrl = 'index.php?option=com_loginguard&task=method.';

			if ($id)
			{
				$nonSefUrl .= 'edit&id=' . (int) $id;
			}
			else
			{
				$nonSefUrl .= 'add&method=' . $record->method;
			}

			$nonSefUrl .= '&user_id=' . $user_id;

			if (!empty($returnURL))
			{
				$nonSefUrl .= '&returnurl=' . urlencode($returnURL);
			}

			$url = Route::_($nonSefUrl, false);
			$this->setRedirect($url, $e->getMessage(), 'error');

			return $this;
		}

		// Update the record's options with the plugin response
		$title = $this->input->getString('title', null);
		$title = trim($title);

		if (empty($title))
		{
			$method = $model->getMethod($record->method);
			$title  = $method['display'];
		}

		// Update the record's "default" flag
		$default         = $this->input->getBool('default', false);
		$record->title   = $title;
		$record->options = $result;
		$record->default = $default ? 1 : 0;

		// Ask the model to save the record
		$saved = $record->store();

		if (!$saved)
		{
			// Go back to the edit page
			$nonSefUrl = 'index.php?option=com_loginguard&task=method.';

			if ($id)
			{
				$nonSefUrl .= 'edit&id=' . (int) $id;
			}
			else
			{
				$nonSefUrl .= 'add';
			}

			$nonSefUrl .= '&user_id=' . $user_id;

			if (!empty($returnURL))
			{
				$nonSefUrl .= '&returnurl=' . urlencode($returnURL);
			}

			$url = Route::_($nonSefUrl, false);
			$this->setRedirect($url, $record->getError(), 'error');

			return $this;
		}

		$this->setRedirect($url);

		return $this;
	}

	/**
	 * Assert that the provided ID is a valid record identified for the given user
	 *
	 * @param   int        $id    Record ID to check
	 * @param   User|null  $user  User record. Null to use current user.
	 *
	 * @return  TfaTable  The loaded record
	 *
	 */
	private function _assertValidRecordId($id, ?User $user = null)
	{
		if (is_null($user))
		{
			$user = $this->app->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);
		}

		/** @var MethodModel $model */
		$model = $this->getModel('Method', 'Administrator');

		$model->setState('id', $id);

		$record = $model->getRecord($user);

		if (is_null($record) || ($record->id != $id) || ($record->user_id != $user->id))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		return $record;
	}

	/**
	 * Assert that the user is logged in.
	 *
	 * @param   User  $user  User record. Null to use current user.
	 *
	 * @throws  RuntimeException  When the user is a guest (not logged in)
	 */
	private function _assertCanEdit(User $user = null)
	{
		if (is_null($user))
		{
			$user = $this->app->getIdentity() ?: Factory::getUser();
		}

		if (!MfaHelper::canEditUser($user))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}
	}

	/**
	 * Assert that the specified MFA Method exists, is activated and enabled for the current user
	 *
	 * @param   string|null  $method  The Method to check
	 *
	 * @return  void
	 *
	 */
	private function _assertMethodExists(?string $method)
	{
		/** @var MethodModel $model */
		$model = $this->getModel('Method', 'Administrator');

		if (empty($method) || !$model->methodExists($method))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}
	}

	private function assertLoggedInUser()
	{
		$user = $this->app->getIdentity()
			?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);

		if ($user->guest)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}
	}
}