<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\Model;

// Prevent direct access
defined('_JEXEC') or die;

use Exception;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Encrypt\Aes;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\Database\ParameterType;

/**
 * A model responsible for converting Joomla's Two Factor Authenticator entries into Akeeba LoginGuard's Multi-factor
 * Authentication entries.
 *
 * @since 1.2.1
 */
class ConvertModel extends BaseDatabaseModel
{
	/**
	 * The site's secret key
	 *
	 * @var   string
	 * @since 1.2.1
	 */
	protected $secret = '';

	/**
	 * @param   int   $limit   How many records to process at once. Around 25 should be safe in most cases.
	 *
	 * @return  bool  True if we converted any Users, false if we're done converting Users
	 * @since   1.2.1
	 */
	public function convert($limit = 25)
	{
		// Get the Users with Joomla! MFA records
		$db     = $this->getDbo();
		$empty1 = '';
		$empty2 = '';
		$query  = $db->getQuery(true)
			->select('*')
			->from($db->qn('#__users'))
			->where($db->qn('otpKey') . ' != :empty1')
			->where($db->qn('otep') . ' != :empty2')
			->bind(':empty1', $empty1)
			->bind(':empty2', $empty2);
		$users  = $db->setQuery($query, 0, $limit)->loadObjectList();

		// There are no more Users with MFA configured, let's stop here
		if (empty($users))
		{
			return false;
		}

		// Get the Secret from Joomla's configuration. It's used as an encryption key in Joomla! 3.6.3 and earlier.
		$secret = $this->getSecret();

		// Loop all Users with MFA
		foreach ($users as $user)
		{
			[$otpMethod, $otpKey] = explode(':', $user->otpKey, 2);
			$otpKey     = $this->decryptTFAString($secret, $otpKey);
			$otep       = $this->decryptTFAString($secret, $user->otep);
			$methodName = 'Convert' . ucfirst($otpMethod);

			// Make sure I can Convert the currently selected Method
			if (!method_exists($this, $methodName))
			{
				continue;
			}

			// Convert the MFA Method
			call_user_func([$this, $methodName], $otpKey, $user->id);

			// Convert the emergency codes
			$this->convertEmergencyCodes($otep, $user->id);

			// Disable the MFA in the user's record
			$update = (object) [
				'id'     => $user->id,
				'otpKey' => '',
				'otep'   => '',
			];
			$db->updateObject('#__users', $update, ['id']);
		}

		return true;
	}

	/**
	 * Private getter for the secret key. If no secret key is set in the model the site's secret key is returned
	 * instead.
	 *
	 * @return  string
	 * @since   1.2.1
	 */
	private function getSecret()
	{
		if (empty($this->secret))
		{
			$this->secret = Factory::getApplication()->get('secret', null);
		}

		return $this->secret;
	}

	/**
	 * Tries to decrypt the MFA configuration, using a different Method depending on the Joomla! version.
	 *
	 * @param   string   $secret            Site's secret key
	 * @param   string   $stringToDecrypt   Base64-encoded and encrypted, JSON-encoded information
	 *
	 * @return  string  Decrypted, but JSON-encoded, information
	 *
	 * @see     https://github.com/joomla/joomla-cms/pull/12497
	 * @since   1.2.1
	 */
	private function decryptTFAString(string $secret, string $stringToDecrypt): string
	{
		// Is this already decrypted?
		try
		{
			$decrypted = json_decode($stringToDecrypt, true);
		}
		catch (Exception $e)
		{
			$decrypted = null;
		}

		if (!empty($decrypted))
		{
			return $stringToDecrypt;
		}

		// No, we need to decrypt the string
		$aes       = new Aes($secret, 256);
		$decrypted = $aes->decryptString($stringToDecrypt);

		if (!is_string($decrypted) || empty($decrypted))
		{
			$aes->setPassword($secret, true);

			$decrypted = $aes->decryptString($stringToDecrypt);
		}

		if (!is_string($decrypted) || empty($decrypted))
		{
			return '';
		}

		// Remove the null padding added during encryption
		return rtrim($decrypted, "\0");
	}

	/**
	 * Convert the one time emergency codes from Joomla's MFA configuration.
	 *
	 * @param   string   $json      The JSON-encoded list of codes
	 * @param   int      $user_id   The ID of the user for this Method
	 *
	 * @return  void
	 * @since   1.2.1
	 */
	private function convertEmergencyCodes($json, $user_id)
	{
		// Try to decode the configuration
		$config = @json_decode($json, true);

		// If the configuration cannot be decoded (corrupt) just give up and don't Convert
		if (empty($config))
		{
			return;
		}

		// Delete any other record with the same user_id and Method.
		$db     = $this->getDbo();
		$method = 'emergencycodes';
		$query  = $db->getQuery(true)
			->delete($db->qn('#__loginguard_tfa'))
			->where($db->qn('user_id') . ' = :user_id')
			->where($db->qn('method') . ' = :method')
			->bind(':user_id', $user_id, ParameterType::INTEGER)
			->bind(':method', $method);
		$db->setQuery($query)->execute();

		// Migrate data
		$jDate = Date::getInstance();
		$table = $this->getTable('Tfa', 'Administrator');

		$table->save([
			'user_id'    => $user_id,
			'title'      => Text::_('COM_LOGINGUARD_LBL_BACKUPCODES'),
			'method'     => 'backupcodes',
			'default'    => 0,
			'created_on' => $jDate->toSql(),
			'last_used'  => $db->getNullDate(),
			'options'    => json_decode($json, true),
		]);
	}

	/**
	 * Public setter for the secret key.
	 *
	 * @param   string   $secret   The secret key to set in the model.
	 *
	 * @return  void
	 * @since   1.2.1
	 */
	public function setSecret($secret)
	{
		$this->secret = $secret;
	}

	/**
	 * Disable all Two Factor Authentication plugins. Since Akeeba LoginGuard will be implementing Multi-factor
	 * Authentication you don't need them any more.
	 *
	 * @return  void
	 * @since   1.2.1
	 */
	public function disableTFA()
	{
		$db     = $this->getDbo();
		$zero   = 0;
		$type   = 'plugin';
		$folder = 'twofactorauth';
		$query  = $db->getQuery(true)
			->update($db->qn('#__extensions'))
			->set($db->qn('enabled') . ' = :zero')
			->where($db->qn('type') . ' = :type')
			->where($db->qn('folder') . ' = :folder')
			->bind(':zero', $zero, ParameterType::INTEGER)
			->bind(':type', $type, ParameterType::STRING)
			->bind(':folder', $folder, ParameterType::STRING);

		$db->setQuery($query)->execute();
	}

	/**
	 * Convert from Joomla MFA Method 'yubikeytotp'. This is a non-standard Method, part of the now defunct Akeeba MFA
	 * plugins for Joomla! 3. It implements MFA with TOTP and/or any number of YubiKeys. We map it to both our TOTP and
	 * our YubiKey plugins.
	 *
	 * @param   string   $json      The JSON-encoded configuration of this Method.
	 * @param   int      $user_id   The ID of the user for this Method
	 *
	 * @return  void
	 * @since   1.2.1
	 */
	private function convertYubikeytotp($json, $user_id)
	{
		$config = @json_decode($json, true);

		if (empty($config))
		{
			return;
		}

		// Handle the TOTP part if there's a non-empty TOTP key code
		if (isset($config['yubikeytotp_code']) && !empty($config['yubikeytotp_code']))
		{
			$fakeConfig = json_encode([
				'code' => $config['yubikeytotp_code'],
			]);

			$this->convertTotp($fakeConfig, $user_id);
		}

		// Handle the YubiKey part if there is an array of YubiKeys in use
		if (isset($config['yubikeytotp']) && is_array($config['yubikeytotp']) && count($config['yubikeytotp']))
		{
			foreach ($config['yubikeytotp'] as $yubikey)
			{
				$fakeConfig = json_encode([
					'yubikey' => $yubikey,
				]);

				$this->convertYubikey($fakeConfig, $user_id);
			}
		}
	}

	/**
	 * Convert from Joomla MFA Method 'totp'. This Method implements MFA with time-based one time passwords, therefore
	 * it's mapped to our TOTP plugin.
	 *
	 * @param   string   $json      The JSON-encoded configuration of this Method.
	 * @param   int      $user_id   The ID of the user for this Method
	 *
	 * @return  void
	 * @since   1.2.1
	 */
	private function convertTotp($json, $user_id)
	{
		$config = @json_decode($json, true);

		if (empty($config))
		{
			return;
		}

		// Delete any other record with the same user_id and Method.
		$db     = $this->getDbo();
		$method = 'totp';
		$query  = $db->getQuery(true)
			->delete($db->qn('#__loginguard_tfa'))
			->where($db->qn('user_id') . ' = :user_id')
			->where($db->qn('method') . ' = :method')
			->bind(':user_id', $user_id, ParameterType::INTEGER)
			->bind(':method', $method, ParameterType::STRING);
		$db->setQuery($query)->execute();

		// Get the Multi-factor Authentication object to insert
		$jDate = Date::getInstance();
		$table = $this->getTable('Tfa', 'Administrator');

		$table->save([
			'user_id'    => $user_id,
			'title'      => 'Authenticator',
			'method'     => 'totp',
			'default'    => 0,
			'created_on' => $jDate->toSql(),
			'last_used'  => $db->getNullDate(),
			'options'    => ['key' => $config['code']],
		]);
	}

	/**
	 * Convert from Joomla MFA Method 'yubikey'. This Method implements MFA with YubiKey only, therefore it's mapped to
	 * our YubiKey plugin.
	 *
	 * @param   string   $json      The JSON-encoded configuration of this Method.
	 * @param   int      $user_id   The ID of the user for this Method
	 *
	 * @return  void
	 * @since   1.2.1
	 */
	private function convertYubikey($json, $user_id)
	{
		// Try to decode the configuration
		$config = @json_decode($json, true);

		// If the configuration cannot be decoded (corrupt) just give up and don't Convert
		if (empty($config))
		{
			return;
		}

		// Delete any other record with the same user_id and Method.
		$db     = $this->getDbo();
		$method = 'yubikey';
		$query  = $db->getQuery(true)
			->delete($db->qn('#__loginguard_tfa'))
			->where($db->qn('user_id') . ' = :user_id')
			->where($db->qn('method') . ' = :method')
			->bind(':user_id', $user_id, ParameterType::INTEGER)
			->bind(':method', $method, ParameterType::STRING);
		$db->setQuery($query)->execute();

		// Migrate data
		$jDate = Date::getInstance();
		$table = $this->getTable('Tfa', 'Administrator');

		$table->save([
			'user_id'    => $user_id,
			'title'      => 'YubiKey ' . $config['yubikey'],
			'method'     => 'yubikey',
			'default'    => 0,
			'created_on' => $jDate->toSql(),
			'last_used'  => $db->getNullDate(),
			'options'    => ['id' => $config['yubikey']],
		]);
	}
}