<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\View\Captive;

// Prevent direct access
defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\Event\BeforeDisplayMethods;
use Akeeba\Component\LoginGuard\Administrator\Event\NotifyActionLog;
use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Model\BackupcodesModel;
use Akeeba\Component\LoginGuard\Administrator\Model\CaptiveModel;
use Exception;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\View\HtmlView as BaseHtmlView;
use Joomla\CMS\Toolbar\ToolbarHelper;
use Joomla\CMS\User\UserFactoryInterface;
use stdClass;

class HtmlView extends BaseHtmlView
{
	/**
	 * The MFA Method records for the current user which correspond to enabled plugins
	 *
	 * @var  array
	 */
	public $records = [];

	/**
	 * The currently selected MFA Method record against which we'll be authenticating
	 *
	 * @var  null|stdClass
	 */
	public $record = null;

	/**
	 * The Captive MFA page's rendering options
	 *
	 * @var   array|null
	 */
	public $renderOptions = null;

	/**
	 * The title to display at the top of the page
	 *
	 * @var   string
	 */
	public $title = '';

	/**
	 * Is this an administrator page?
	 *
	 * @var   bool
	 */
	public $isAdmin = false;

	/**
	 * Does the currently selected Method allow authenticating against all of its records?
	 *
	 * @var   bool
	 */
	public $allowEntryBatching = false;

	/**
	 * All enabled MFA Methods (plugins)
	 *
	 * @var   array
	 */
	public $mfaMethods;

	/**
	 * Browser identification hash (fingerprint)
	 *
	 * @var   string|null
	 * @since 3.3.0
	 */
	public $browserId;

	/**
	 * Execute and display a template script.
	 *
	 * @param   string  $tpl  The name of the template file to parse; automatically searches through the template paths.
	 *
	 * @return  void  A string if successful, otherwise an Error object.
	 *
	 * @throws  Exception
	 */
	public function display($tpl = null)
	{
		$app  = Factory::getApplication();
		$user = Factory::getApplication()->getIdentity()
			?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);

		MfaHelper::triggerEvent(new BeforeDisplayMethods($user));

		/** @var CaptiveModel $model */
		$model = $this->getModel();

		// Load data from the model
		$this->isAdmin    = $app->isClient('administrator');
		$this->records    = $this->get('records');
		$this->record     = $this->get('record');
		$this->mfaMethods = MfaHelper::getMfaMethods();
		$this->browserId  = $app->getSession()->get('com_loginguard.browserId', null);

		if (!empty($this->records))
		{
			/** @var BackupcodesModel $codesModel */
			$codesModel        = $this->getModel('Backupcodes');
			$backupCodesRecord = $codesModel->getBackupCodesRecord();

			if (!is_null($backupCodesRecord))
			{
				$backupCodesRecord->title = Text::_('COM_LOGINGUARD_LBL_BACKUPCODES');
				$this->records[]          = $backupCodesRecord;
			}
		}

		// If we only have one record there's no point asking the user to select a MFA Method
		if (empty($this->record) && !empty($this->records))
		{
			// Default to the first record
			$this->record = reset($this->records);

			// If we have multiple records try to make this record the default
			if (count($this->records) > 1)
			{
				foreach ($this->records as $record)
				{
					if ($record->default)
					{
						$this->record = $record;

						break;
					}
				}
			}
		}

		// Set the correct layout based on the availability of a MFA record
		$this->setLayout('default');

		// Should I implement the Remember Me feature?
		$rememberMe = ComponentHelper::getParams('com_loginguard')->get('allow_rememberme', 1);

		// If we have no record selected or explicitly asked to run the 'select' task use the correct layout
		if (is_null($this->record) || ($model->getState('task') == 'select'))
		{
			$this->setLayout('select');
		}
		// If there's no browser ID try to fingerprint the browser instead of showing the MFA page
		elseif (is_null($this->browserId) && ($rememberMe == 1))
		{
			$this->setLayout('fingerprint');
		}

		switch ($this->getLayout())
		{
			case 'select':
				$this->allowEntryBatching = 1;

				MfaHelper::triggerEvent(
					new NotifyActionLog('onComLoginguardCaptiveShowSelect', [])
				);
				break;

			case 'fingerprint':
				// This flag tells the Captive model that we are sending a new browser ID now
				$app->getSession()->set('com_loginguard.browserIdCodeLoaded', true);
				break;

			case 'default':
			default:
				$this->renderOptions      = $model->loadCaptiveRenderOptions($this->record);
				$this->allowEntryBatching = $this->renderOptions['allowEntryBatching'] ?? 0;

				MfaHelper::triggerEvent(
					new NotifyActionLog('onComLoginguardCaptiveShowCaptive', [
						$this->escape($this->record->title),
					])
				);
				break;
		}

		// Which title should I use for the page?
		$this->title = $this->get('PageTitle');

		// Back-end: always show a title in the 'title' module position, not in the page body
		if ($this->isAdmin)
		{
			ToolbarHelper::title(Text::_('COM_LOGINGUARD_HEAD_MFA_PAGE'), 'loginguard');
			$this->title = '';
		}

		// Include CSS
		$wam = $this->document->getWebAssetManager();
		$wam->useStyle('com_loginguard.captive');

		if (ComponentHelper::getParams('com_loginguard')->get('dark_mode') != 0)
		{
			$wam->useStyle('com_loginguard.dark');
		}

		// Display the view
		parent::display($tpl);
	}
}