<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Loginguard\Email\Extension;

// Prevent direct access
defined('_JEXEC') || die;

use Akeeba\Component\LoginGuard\Administrator\DataShape\CaptiveRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\DataShape\MethodDescriptor;
use Akeeba\Component\LoginGuard\Administrator\DataShape\SetupRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\Event\BeforeDisplayMethods;
use Akeeba\Component\LoginGuard\Administrator\Event\GetInfo;
use Akeeba\Component\LoginGuard\Administrator\Event\SaveSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowCaptive;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\Validate;
use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Encrypt\Totp;
use Joomla\CMS\Factory;
use Joomla\CMS\Input\Input;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;
use RuntimeException;
use function count;

/**
 * Akeeba LoginGuard Plugin for Multi-factor Authentication Method "Authentication Code by PushBullet"
 *
 * Requires entering a 6-digit code sent to the user through email. These codes change automatically every 30
 * seconds.
 */
class Email extends CMSPlugin implements SubscriberInterface
{
	/**
	 * Generated OTP length. Constant: 6 numeric digits.
	 */
	private const CODE_LENGTH = 6;

	/**
	 * Length of the secret key used for generating the OTPs. Constant: 10 characters.
	 */
	private const SECRET_KEY_LENGTH = 20;

	/**
	 * The CMS application we are running under
	 *
	 * @var   CMSApplication
	 * @since 6.0.0
	 */
	protected $app;

	/**
	 * Forbid registration of legacy (Joomla 3) event listeners.
	 *
	 * @var    boolean
	 * @since  6.0.0
	 *
	 * @deprecated
	 */
	protected $allowLegacyListeners = false;

	/**
	 * Autoload this plugin's language files
	 *
	 * @var    boolean
	 * @since  6.0.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * The MFA Method name handled by this plugin
	 *
	 * @var   string
	 */
	private $mfaMethodName = 'email';

	/**
	 * Returns an array of events this subscriber will listen to.
	 *
	 * @return  array
	 *
	 * @since   6.0.0
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'onLoginGuardMfaGetMethod'         => 'onLoginGuardMfaGetMethod',
			'onLoginGuardMfaCaptive'           => 'onLoginGuardMfaCaptive',
			'onLoginGuardMfaGetSetup'          => 'onLoginGuardMfaGetSetup',
			'onLoginGuardMfaSaveSetup'         => 'onLoginGuardMfaSaveSetup',
			'onLoginGuardMfaValidate'          => 'onLoginGuardMfaValidate',
			'onLoginGuardBeforeDisplayMethods' => 'onLoginGuardBeforeDisplayMethods',
		];
	}

	/**
	 * Gets the identity of this MFA Method
	 *
	 * @param   GetInfo  $event  The event we are handling
	 *
	 * @return  void
	 */
	public function onLoginGuardMfaGetMethod(GetInfo $event): void
	{
		$event->addResult(
			new MethodDescriptor([
				'name'      => $this->mfaMethodName,
				'display'   => Text::_('PLG_LOGINGUARD_EMAIL_LBL_DISPLAYEDAS'),
				'shortinfo' => Text::_('PLG_LOGINGUARD_EMAIL_LBL_SHORTINFO'),
				'image'     => 'media/plg_loginguard_email/images/email.svg',
				'help_url'  => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Email'),
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the Captive MFA page. This is the page which appears
	 * right after you log in and asks you to validate your login with MFA.
	 *
	 * @param   ShowCaptive  $event  The event we are handling
	 *
	 * @return  void
	 */
	public function onLoginGuardMfaCaptive(ShowCaptive $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$key     = $options['key'] ?? '';
		// Send an email message with a new code and ask the user to enter it.
		$user = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($record->user_id);

		try
		{
			$this->sendCode($key, $user);
		}
		catch (Exception $e)
		{
			return;
		}

		$event->addResult(
			new CaptiveRenderOptions([
				// Custom HTML to display above the MFA form
				'pre_message'        => '',
				// How to render the MFA code field. "input" (HTML input element) or "custom" (custom HTML)
				'field_type'         => 'input',
				// The type attribute for the HTML input box. Typically "text" or "password". Use any HTML5 input type.
				'input_type'         => 'number',
				// Placeholder text for the HTML input box. Leave empty if you don't need it.
				'placeholder'        => Text::_('PLG_LOGINGUARD_EMAIL_LBL_SETUP_PLACEHOLDER'),
				// Label to show above the HTML input box. Leave empty if you don't need it.
				'label'              => Text::_('PLG_LOGINGUARD_EMAIL_LBL_SETUP_LABEL'),
				// Custom HTML. Only used when field_type = custom.
				'html'               => '',
				// Custom HTML to display below the MFA form
				'post_message'       => '',
				// Should I hide the default Submit button?
				'hide_submit'        => false,
				// Is this MFA method validating against all configured authenticators of the same type?
				'allowEntryBatching' => false,
				// URL for help content
				'help_url'           => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Email'),
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the MFA setup page. This is the page which allows the
	 * user to add or modify a MFA Method for their user account. If the record does not correspond to your plugin
	 * return an empty array.
	 *
	 * @param   ShowSetup  $event
	 *
	 * @return  void
	 */
	public function onLoginGuardMfaGetSetup(ShowSetup $event): void
	{
		/** @var TfaTable $record The #__loginguard_tfa record currently selected by the user. */
		$record = $event['record'];

		$helpURL = $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Email');

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options           = $this->_decodeRecordOptions($record);
		$key               = $options['key'] ?? '';
		$isKeyAlreadySetup = !empty($key);

		// If there's a key in the session use that instead.
		$session = $this->app->getSession();
		$session->get('com_loginguard.emailcode.key', $key);

		// Initialize objects
		$timeStep = min(max((int) $this->params->get('timestep', 120), 30), 900);
		$totp     = new Totp($timeStep, self::CODE_LENGTH, self::SECRET_KEY_LENGTH);

		// If there's still no key in the options, generate one and save it in the session
		if (!$isKeyAlreadySetup)
		{
			$key = $totp->generateSecret();

			$session->set('com_loginguard.emailcode.key', $key);
			$session->set('com_loginguard.emailcode.user_id', $record->user_id);

			$user = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($record->user_id);

			$this->sendCode($key, $user);

			$event->addResult(
				new SetupRenderOptions([
					'default_title' => Text::_('PLG_LOGINGUARD_EMAIL_LBL_DISPLAYEDAS'),
					'hidden_data'   => [
						'key' => $key,
					],
					'field_type'    => 'input',
					'input_type'    => 'number',
					'input_value'   => '',
					'placeholder'   => Text::_('PLG_LOGINGUARD_EMAIL_LBL_SETUP_PLACEHOLDER'),
					'label'         => Text::_('PLG_LOGINGUARD_EMAIL_LBL_SETUP_LABEL'),
					'help_url'      => $helpURL,
				])
			);
		}
		else
		{
			$event->addResult(
				new SetupRenderOptions([
					'default_title' => Text::_('PLG_LOGINGUARD_EMAIL_LBL_DISPLAYEDAS'),
					'field_type'    => 'custom',
					'html'          => '',
					'help_url'      => $helpURL,
				])
			);
		}
	}

	/**
	 * Parse the input from the MFA setup page and return the configuration information to be saved to the database. If
	 * the information is invalid throw a RuntimeException to signal the need to display the editor page again. The
	 * message of the exception will be displayed to the user. If the record does not correspond to your plugin return
	 * an empty array.
	 *
	 * @param   SaveSetup  $event
	 *
	 * @return  void
	 */
	public function onLoginGuardMfaSaveSetup(SaveSetup $event): void
	{
		/**
		 * @var TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 * @var Input    $input  The user input you are going to take into account.
		 */
		$record = $event['record'];
		$input  = $event['input'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			return;
		}

		// Load the options from the record (if any)
		$options           = $this->_decodeRecordOptions($record);
		$key               = $options['key'] ?? '';
		$isKeyAlreadySetup = !empty($key);
		$session           = $this->app->getSession();

		// If there is no key in the options fetch one from the session
		if (empty($key))
		{
			$key = $session->get('com_loginguard.emailcode.key', null);
		}

		// If there is still no key in the options throw an error
		if (empty($key))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		/**
		 * If the code is empty but the key already existed in $options someone is simply changing the title / default
		 * Method status. We can allow this and stop checking anything else now.
		 */
		$code = $input->getCmd('code');

		if (empty($code) && $isKeyAlreadySetup)
		{
			$event->addResult($options);

			return;
		}

		// In any other case validate the submitted code
		$timeStep = min(max((int) $this->params->get('timestep', 120), 30), 900);
		$totp     = new Totp($timeStep, self::CODE_LENGTH, self::SECRET_KEY_LENGTH);
		$isValid  = $totp->checkCode((string) $key, (string) $code);

		if (!$isValid)
		{
			throw new RuntimeException(Text::_('PLG_LOGINGUARD_EMAIL_ERR_INVALID_CODE'), 500);
		}

		// The code is valid. Unset the key from the session.
		$session->set('com_loginguard.emailcode.key', null);

		// Return the configuration to be serialized
		$event->addResult([
			'key' => $key,
		]);
	}

	/**
	 * Validates the Two Factor Authentication code submitted by the user in the Captive Multi-factor Authentication
	 * page. If the record does not correspond to your plugin return FALSE.
	 *
	 * @param   Validate  $event
	 *
	 * @return  void
	 */
	public function onLoginGuardMfaValidate(Validate $event): void
	{
		/**
		 * @var   TfaTable    $record The MFA Method's record you're validating against
		 * @var   User        $user   The user record
		 * @var   string|null $code   The submitted code
		 */
		$record = $event['record'];
		$user   = $event['user'];
		$code   = $event['code'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			$event->addResult(false);

			return;
		}

		// Double check the MFA Method is for the correct user
		if ($user->id != $record->user_id)
		{
			$event->addResult(false);

			return;
		}

		// Load the options from the record (if any)
		$options = $this->_decodeRecordOptions($record);
		$key     = $options['key'] ?? '';

		// If there is no key in the options throw an error
		if (empty($key))
		{
			$event->addResult(false);

			return;
		}

		// Check the MFA code for validity
		$timeStep = min(max((int) $this->params->get('timestep', 120), 30), 900);
		$totp     = new Totp($timeStep, self::CODE_LENGTH, self::SECRET_KEY_LENGTH);

		$event->addResult($totp->checkCode((string) $key, (string) $code));
	}

	/**
	 * Executes before showing the MFA Methods for the user. Used for the Force Enable feature.
	 *
	 * @param   BeforeDisplayMethods  $event
	 *
	 * @return  void
	 * @throws Exception
	 */
	public function onLoginGuardBeforeDisplayMethods(BeforeDisplayMethods $event): void
	{
		/** @var ?User $user */
		$user = $event['user'];

		// Is the forced enable feature activated?
		if ($this->params->get('force_enable', 0) != 1)
		{
			return;
		}

		// Get second factor Methods for this user
		$userMfaRecords = MfaHelper::getUserMfaRecords($user->id);

		// If there are no Methods go back
		if (count($userMfaRecords) < 1)
		{
			return;
		}

		// If the only Method is backup codes go back
		if (count($userMfaRecords) == 1)
		{
			/** @var TfaTable $record */
			$record = reset($userMfaRecords);

			if ($record->method == 'backupcodes')
			{
				return;
			}
		}

		// If I already have the email Method go back
		$emailRecords = array_filter($userMfaRecords, function (TfaTable $record) {
			return $record->method == 'email';
		});

		if (count($emailRecords))
		{
			return;
		}

		// Add the email Method
		try
		{
			/** @var MVCFactoryInterface $factory */
			$factory = $this->app->bootComponent('com_loginguard')->getMVCFactory();
			/** @var TfaTable $record */
			$record = $factory->createTable('Tfa', 'Administrator');
			$record->reset();

			$timeStep = min(max((int) $this->params->get('timestep', 120), 30), 900);
			$totp     = new Totp($timeStep, self::CODE_LENGTH, self::SECRET_KEY_LENGTH);

			$record->save([
				'method'  => 'email',
				'title'   => Text::_('PLG_LOGINGUARD_EMAIL_LBL_DISPLAYEDAS'),
				'options' => [
					'key' => ($totp)->generateSecret(),
				],
				'default' => 0,
			]);
		}
		catch (Exception $event)
		{
			// Fail gracefully
		}
	}

	/**
	 * Creates a new TOTP code based on secret key $key and sends it to the user via email.
	 *
	 * @param   string     $key   The TOTP secret key
	 * @param   User|null  $user  The Joomla! user to use
	 *
	 * @return  void
	 *
	 */
	private function sendCode(string $key, ?User $user = null)
	{
		static $alreadySent = false;

		// Make sure we have a user
		if (!is_object($user) || !($user instanceof User))
		{
			$user = $this->app->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);
		}

		// Get the API objects
		$timeStep = min(max((int) $this->params->get('timestep', 120), 30), 900);
		$totp     = new Totp($timeStep, self::CODE_LENGTH, self::SECRET_KEY_LENGTH);

		// Create the list of variable replacements
		$code = $totp->getCode($key);

		$replacements = [
			'[CODE]'     => $code,
			'[SITENAME]' => $this->app->get('sitename'),
			'[SITEURL]'  => Uri::base(),
			'[USERNAME]' => $user->username,
			'[EMAIL]'    => $user->email,
			'[FULLNAME]' => $user->name,
		];

		// Get the title and body of the e-mail message
		$subject = Text::_('PLG_LOGINGUARD_EMAIL_MESSAGE_SUBJECT');
		$subject = str_ireplace(array_keys($replacements), array_values($replacements), $subject);
		$body    = Text::_('PLG_LOGINGUARD_EMAIL_MESSAGE_BODY');
		$body    = str_ireplace(array_keys($replacements), array_values($replacements), $body);

		if ($alreadySent)
		{
			return;
		}

		$alreadySent = true;

		// Send email
		try
		{
			$mailer = Factory::getMailer();
			$mailer->setSubject($subject);
			$mailer->setBody($body);
			$mailer->addRecipient($user->email, $user->name);
			$mailer->Send();
		}
		catch (Exception $e)
		{
			return;
		}
	}

	/**
	 * Decodes the options from a #__loginguard_tfa record into an options object.
	 *
	 * @param   TfaTable  $record
	 *
	 * @return  array
	 */
	private function _decodeRecordOptions(TfaTable $record): array
	{
		$options = [
			'key' => '',
		];

		if (!empty($record->options))
		{
			$recordOptions = $record->options;

			$options = array_merge($options, $recordOptions);
		}

		return $options;
	}
}
