<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\System\Loginguard\Extension;

defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\Database\DatabaseDriver;
use Joomla\Event\DispatcherInterface;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;

class LoginGuard extends CMSPlugin implements SubscriberInterface
{
	/**
	 * Application object.
	 *
	 * @var    CMSApplication
	 * @since  5.0.0
	 */
	protected $app;

	/**
	 * Database object.
	 *
	 * @var    DatabaseDriver
	 * @since  5.0.0
	 */
	protected $db;

	/**
	 * Forbid registration of legacy (Joomla 3) event listeners.
	 *
	 * @var    boolean
	 * @since  6.0.0
	 *
	 * @deprecated
	 */
	protected $allowLegacyListeners = false;

	/**
	 * User groups for which Multi-factor Authentication is never applied
	 *
	 * @var   array
	 * @since 3.0.1
	 */
	private $neverMFAUserGroups = [];

	/**
	 * User groups for which Multi-factor Authentication is mandatory
	 *
	 * @var   array
	 * @since 3.0.1
	 */
	private $forceMFAUserGroups = [];

	/**
	 * Constructor
	 *
	 * @param   DispatcherInterface  &$subject   The object to observe
	 * @param   array                 $config    An optional associative array of configuration settings.
	 *                                           Recognized key values include 'name', 'group', 'params', 'language'
	 *                                           (this list is not meant to be comprehensive).
	 */
	public function __construct(&$subject, $config = [])
	{
		parent::__construct($subject, $config);

		$cParams = ComponentHelper::getParams('com_loginguard');

		$this->neverMFAUserGroups = $cParams->get('neverTSVUserGroups', []);
		$this->neverMFAUserGroups = is_array($this->neverMFAUserGroups) ? $this->neverMFAUserGroups : [];

		$this->forceMFAUserGroups = $cParams->get('forceTSVUserGroups', []);
		$this->forceMFAUserGroups = is_array($this->forceMFAUserGroups) ? $this->forceMFAUserGroups : [];
	}

	/**
	 * Returns an array of events this subscriber will listen to.
	 *
	 * @return  array
	 *
	 * @since   6.0.0
	 */
	public static function getSubscribedEvents(): array
	{
		if (!ComponentHelper::isEnabled('com_loginguard'))
		{
			return [];
		}

		return [
			'onAfterRoute'     => 'onAfterRoute',
			'onUserAfterLogin' => 'onUserAfterLogin',
		];
	}

	/**
	 * Gets triggered right after Joomla has finished with the SEF routing and before it has the chance to dispatch the
	 * application (load any components).
	 *
	 * @param   Event  $event  The event we are handling
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 */
	public function onAfterRoute(Event $event): void
	{
		$input = $this->app->input;

		// We only kick in in the Site and Administrator applications
		if (!$this->app->isClient('site') && !$this->app->isClient('administrator'))
		{
			return;
		}

		// This also checks if we are in the site/administrator app and the user is logged in.
		if (!$this->willNeedRedirect())
		{
			return;
		}

		// If this is non-HTML content and also not com_ajax: access denied
		if ($input->get('format', 'html') !== 'html' && $input->getCmd('option') !== 'com_ajax')
		{
			throw new \RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Make sure we are logged in
		try
		{
			$user = $this->app->getIdentity();
		}
		catch (Exception $e)
		{
			return;
		}

		/**
		 * If you have enabled Joomla's Privacy Consent you'd end up with an infinite redirection loop. That's because
		 * Joomla! did a partial copy of my original research code on Captive Joomla! logins. They did not implement
		 * configurable exceptions since they do not know or care about third party extensions -- even when it's the
		 * same extensions they copied code from.
		 *
		 * While on Joomla 3 we can snuff out the onAfterRoute event handler for the privacy plugin, this is not
		 * possible on Joomla 4 because all event handlers are Closure objects which cannot be inspected with Reflection
		 * (because screw you, PHP developer, that's why!) so we have to instead disable the redirection IF we are on
		 * Joomla 4 AND the privacy plugin is enabled AND the user hasn't consented yet.
		 */
		if (
			PluginHelper::isEnabled('system', 'privacyconsent')
			&& !$this->isUserConsented((int) $user->id)
		)
		{
			return;
		}

		// We only kick in when the user has actually set up MFA or must definitely enable MFA.
		$needsMFA     = $this->needsMFA($user);
		$disabledMFA  = $this->disabledMFA($user);
		$mandatoryMFA = $this->mandatoryMFA($user);
		$session      = $this->app->getSession();

		if ($needsMFA && !$disabledMFA)
		{
			/**
			 * Saves the current URL as the return URL if all of the following conditions apply
			 * - It is not a URL to LoginGuard itself
			 * - A return URL does not already exist, is imperfect or external to the site
			 *
			 * If no return URL has been set up and the current URL is LoginGuard itself we will save the home page as
			 * the redirect target.
			 */
			$returnUrl       = $session->get('com_loginguard.return_url', '');
			$imperfectReturn = $session->get('com_loginguard.imperfect_return', false);
			$option          = $input->getCmd('option', null);

			if ($imperfectReturn || empty($returnUrl) || !Uri::isInternal($returnUrl))
			{
				if ($option != 'com_loginguard')
				{
					$session->set(
						'com_loginguard.return_url', Uri::getInstance()->toString([
						'scheme', 'user', 'pass', 'host', 'port', 'path', 'query', 'fragment',
					])
					);
				}
				elseif (empty($returnUrl))
				{
					$session->set('com_loginguard.return_url', Uri::base());
				}
			}

			// Redirect
			$captiveUrl = $session->get('com_loginguard.captiveUrl')
				?:
				Route::_('index.php?option=com_loginguard&view=captive', false);
			$session->set('com_loginguard.captiveUrl', null);

			$this->app->redirect($captiveUrl, 307);

			return;
		}

		// If we're here someone just logged in but does not have MFA set up. Just flag him as logged in and continue.
		$session->set('com_loginguard.mfa_checked', 1);

		// If we don't have MFA set up yet AND the user plugin had set up a redirection we will honour it
		$redirectionUrl = $session->get('com_loginguard.postloginredirect', null);

		// If the user is in a group that requires MFA we will redirect them to the setup page
		if (!$needsMFA && $mandatoryMFA)
		{
			// First unset the flag to make sure the redirection will apply until they conform to the mandatory MFA
			$session->set('com_loginguard.mfa_checked', 0);

			// Now set a flag which forces rechecking MFA for this user
			$session->set('com_loginguard.recheck_mandatory_mfa', 1);

			// Then redirect them to the setup page
			$this->redirectToMFASetup();
		}

		if (!$needsMFA && $redirectionUrl && !$disabledMFA)
		{
			$session->set('com_loginguard.postloginredirect', null);

			$this->app->redirect($redirectionUrl);
		}
	}

	/**
	 * Hooks on the Joomla! login event. Detects silent logins and disables the Multi-factor Authentication Captive page
	 * in this case.
	 *
	 * Moreover, it will save the redirection URL and the Captive URL which is necessary in Joomla 4. You see, in Joomla
	 * 4 having unified sessions turned on makes the backend login redirect you to the frontend of the site AFTER
	 * logging in, something which would cause the Captive page to appear in the frontend and redirect you to the public
	 * frontend homepage after successfully passing the Multi-factor Authentication process.
	 *
	 * @param   Event  $event  The event we are handling
	 */
	public function onUserAfterLogin(Event $event): void
	{
		/**
		 * @var   array $options Passed by Joomla. user: a User object; responseType: string, authentication response type.
		 */
		[$options] = $event->getArguments();
		$session = $this->app->getSession();

		// Always reset the browser ID to avoid session poisoning attacks
		$session->set('com_loginguard.browserId', null);
		$session->set('com_loginguard.browserIdCodeLoaded', false);

		// Save the current URL and mark it as an imperfect return (we'll fall back to it if all else fails)
		$return_url = $session->get('com_loginguard.return_url', '')
			?:
			Uri::getInstance()->toString([
				'scheme', 'user', 'pass', 'host', 'port', 'path', 'query', 'fragment',
			]);
		$session->set('com_loginguard.return_url', $return_url);
		$session->set('com_loginguard.imperfect_return', true);

		// Set up the correct Captive URL
		$captiveUrl = $session->get('com_loginguard.captiveUrl')
			?:
			Route::_('index.php?option=com_loginguard&view=captive', false);
		$session->set('com_loginguard.captiveUrl', $captiveUrl);

		// Should I show MFA even on silent logins? Default: 1 (yes, show)
		if (((int) $this->params->get('2svonsilent', 1)) === 1)
		{
			return;
		}

		// Make sure I have a valid user
		/** @var User $user */
		$user = $options['user'];

		if (!is_object($user) || !($user instanceof User))
		{
			return;
		}

		// Is this a silent login?
		if (!$this->isSilentLogin($user, $options['responseType']))
		{
			return;
		}

		// Set the flag indicating that MFA is already checked.
		$session->set('com_loginguard.mfa_checked', 1);
	}

	/**
	 * Does the current user need to complete MFA authentication before being allowed to access the site?
	 *
	 * @param   User  $user  The user object
	 *
	 * @return  bool
	 */
	private function needsMFA(User $user): bool
	{
		static $records = null;

		// Get the user's MFA records
		$records = $records ?? MfaHelper::getUserMfaRecords($user->id);

		// No MFA Methods? Then we obviously don't need to display a Captive login page.
		if (count($records) < 1)
		{
			return false;
		}

		// Let's get a list of all currently active MFA Methods
		$mfaMethods = MfaHelper::getMfaMethods();

		// If not MFA Method is active we can't really display a Captive login page.
		if (empty($mfaMethods))
		{
			return false;
		}

		// Get a list of just the Method names
		$methodNames = [];

		foreach ($mfaMethods as $mfaMethod)
		{
			$methodNames[] = $mfaMethod->name;
		}

		// Filter the records based on currently active MFA Methods
		return array_reduce(
			$records,
			function (bool $carry, $record) use ($methodNames) {
				return $carry || in_array($record->method, $methodNames);
			},
			false
		);
	}

	/**
	 * Does the user belong in a group indicating MFA should be disabled for them?
	 *
	 * @param   User  $user
	 *
	 * @return  bool
	 */
	private function disabledMFA(User $user): bool
	{
		// If the user belongs to a "never check for MFA" user group they are exempt from MFA
		$userGroups             = $user->getAuthorisedGroups();
		$belongsToMFAUserGroups = array_intersect($this->neverMFAUserGroups, $userGroups);

		return !empty($belongsToMFAUserGroups);
	}

	/**
	 * Does the user belong in a group indicating MFA is required for them?
	 *
	 * @param   User  $user
	 *
	 * @return  bool
	 */
	private function mandatoryMFA(User $user): bool
	{
		// If the user belongs to a "never check for MFA" user group they are exempt from MFA
		$userGroups             = $user->getAuthorisedGroups();
		$belongsToMFAUserGroups = array_intersect($this->forceMFAUserGroups, $userGroups);

		return !empty($belongsToMFAUserGroups);
	}

	/**
	 * Redirect the user to the Multi-factor Authentication Method setup page.
	 *
	 * @return  void
	 *
	 * @since   3.0.1
	 */
	private function redirectToMFASetup(): void
	{
		// If we are in a LoginGuard page do not redirect
		$option = strtolower($this->app->input->getCmd('option'));

		if ($option == 'com_loginguard')
		{
			return;
		}

		// Otherwise redirect to the LoginGuard MFA setup page after enqueueing a message
		$url = Route::_('index.php?option=com_loginguard&view=methods');
		$this->app->redirect($url, 307);
	}

	/**
	 * Check whether we'll need to do a redirection to the Captive page.
	 *
	 * @return  bool
	 *
	 * @throws  Exception
	 * @since   3.0.4
	 *
	 */
	private function willNeedRedirect(): bool
	{
		/**
		 * We only kick in if the session flag is not set AND the user is not flagged for monitoring of their MFA status
		 *
		 * In case a user belongs to a group which requires MFA to be always enabled and they logged in without having
		 * MFA enabled we have the recheck flag. This prevents the user from enabling and immediately disabling MFA,
		 * circumventing the requirement for MFA.
		 */
		$session    = $this->app->getSession();
		$mfaChecked = $session->get('com_loginguard.mfa_checked', 0) != 0;
		$mfaRecheck = $session->get('com_loginguard.recheck_mandatory_mfa', 0) != 0;

		if ($mfaChecked && !$mfaRecheck)
		{
			return false;
		}

		// Make sure we are logged in
		try
		{
			$user = $this->app->getIdentity();
		}
		catch (Exception $e)
		{
			// This would happen if we are in CLI or under an old Joomla! version. Either case is not supported.
			return false;
		}

		// The plugin only needs to kick in when you have logged in
		if (empty($user) || $user->guest)
		{
			return false;
		}

		/**
		 * Special handling when the requireReset flag is set on the user account.
		 *
		 * Joomla checks the requireReset flag on the user account in the application's doExecute Method. If it is set
		 * it will call CMSApplication::checkUserRequireReset() which issues a redirection for the user to reset their
		 * password.
		 *
		 * One easy option here is to say "if the user must reset their password don't show the MFA Captive page"
		 * Unfortunately, that would be a bad idea because of the naive and insecure manner Joomla goes about the forced
		 * password reset. Instead of going through the actual password reset (“Forgot your password?”) page it instead
		 * redirects the user the user profile editor page! This allows the logged in user to view and change everything
		 * in the user profile, including disabling and changing the MFA options. Considering that forced password reset
		 * is meant to be primarily used when we suspect that the user's account has been compromised this creates a
		 * grave security risk. The attacker in possession of the username and password can trick a Super User into
		 * forcing a password reset, thereby allowing them to bypass Multi-factor Authentication and take over the user
		 * account.
		 *
		 * Instead, we unset the requireReset user flag for the duration of the page load when this Method here is
		 * called. This prevents Joomla from redirecting. As a result you need to go through Multi-factor Authentication
		 * as per usual. Once you do that the mfa_checked flag is set in the session and this Method never reaches this
		 * point of execution where we unset the requireReset flag. Therefore Joomla now sees the requireReset flag and
		 * shows you the user profile edit page. Now it's safe to do so since you have already proven your identity by
		 * means of Multi-factor Authentication i.e. there's no doubt we should let you make any kind of user account
		 * change.
		 *
		 * @see \Joomla\CMS\Application\SiteApplication::doExecute()
		 * @see \Joomla\CMS\Application\CMSApplication::checkUserRequireReset()
		 */
		if ($user->requireReset || 0)
		{
			$user->requireReset = 0;
		}


		// If we are in the administrator section we only kick in when the user has backend access privileges
		$isAdmin = $this->app->isClient('administrator');

		if ($isAdmin && !$user->authorise('core.login.admin'))
		{
			return false;
		}

		$needsMFA = $this->needsMFA($user);

		if ($mfaChecked && $mfaRecheck && $needsMFA)
		{
			return false;
		}

		// We only kick in if the option and task are not the ones of the Captive page
		$fallbackView = $this->app->input->getCmd('controller', '');
		$option       = strtolower($this->app->input->getCmd('option', ''));
		$task         = strtolower($this->app->input->getCmd('task', ''));
		$view         = strtolower($this->app->input->getCmd('view', $fallbackView));

		if (strpos($task, '.') !== false)
		{
			$parts = explode('.', $task);
			$view  = ($parts[0] ?? $view) ?: $view;
			$task  = ($parts[1] ?? $task) ?: $task;
		}

		if ($option == 'com_loginguard')
		{
			// In case someone gets any funny ideas...
			$this->app->input->set('tmpl', 'index');
			$this->app->input->set('format', 'html');
			$this->app->input->set('layout', null);

			if (empty($view) && (strpos($task, '.') !== false))
			{
				[$view, $task] = explode('.', $task, 2);
			}

			// The Captive login page is always allowed
			if ($view === 'captive')
			{
				return false;
			}

			// These views are only allowed if you do not have MFA enabled *or* if you have already logged in.
			if (!$needsMFA && in_array($view, ['callback', 'method', 'methods']))
			{
				return false;
			}

			return true;
		}

		// Allow the frontend user to log out (in case they forgot their MFA code or something)
		if (!$isAdmin && ($option == 'com_users') && ($view == 'user') && ($task == 'logout'))
		{
			return false;
		}

		// Allow the backend user to log out (in case they forgot their MFA code or something)
		if ($isAdmin && ($option == 'com_login') && ($task == 'logout'))
		{
			return false;
		}

		/**
		 * Allow com_ajax. This is required for cookie acceptance in the following scenario. Your session has expired,
		 * therefore you need to re-apply MFA. Moreover, your cookie acceptance cookie has also expired and you need to
		 * accept the site's cookies again.
		 */
		if ($option == 'com_ajax')
		{
			return false;
		}

		return true;
	}

	/**
	 * Suppress Multi-factor Authentication when Joomla performs a silent login (cookie, social login / single sign-on, GMail,
	 * LDAP). In these cases the login risk has been managed externally.
	 *
	 * For your reference, the Joomla authentication response types are as follows:
	 *
	 * - Joomla: username and password login. We recommend using MFA with it.
	 * - Cookie: "Remember Me" cookie with a secure, single use token and other safeguards for the user session.
	 * - GMail: login with GMail credentials (probably no longer works)
	 * - LDAP: Joomla's LDAP plugin
	 * - SocialLogin: Akeeba Social Login (login with Facebook etc)
	 *
	 * @param   User    $user
	 * @param   string  $responseType
	 *
	 * @return  bool
	 *
	 * @since   3.1.0
	 */
	private function isSilentLogin(User $user, $responseType): bool
	{
		// Fail early if the user is not properly logged in.
		if (!is_object($user) || $user->guest)
		{
			return false;
		}

		// Get the custom Joomla login responses we will consider "silent"
		$rawCustomResponses = $this->params->get('silentresponses', '');
		$customResponses    = explode(',', $rawCustomResponses);
		$customResponses    = array_map('trim', $customResponses);
		$customResponses    = array_filter($customResponses, function ($x) {
			return !empty($x);
		});
		$silentResponses    = array_unique($customResponses);

		// If all else fails, use our default list (Joomla's Remember Me cookie and Akeeba SocialLogin)
		if (empty($silentResponses))
		{
			$silentResponses = ['cookie', 'sociallogin', 'passwordless'];
		}

		// Is it a silent login after all?
		if (is_string($responseType) && !empty($responseType) && in_array(strtolower($responseType), $silentResponses))
		{
			return true;
		}

		return false;
	}

	/**
	 * Method to check if the given user has consented yet
	 *
	 * @param   integer  $userId  ID of uer to check
	 *
	 * @return  boolean
	 *
	 * @since   5.0.0
	 */
	private function isUserConsented(int $userId): bool
	{
		// Guest user? WTF!
		if (empty($userId))
		{
			return true;
		}

		$userId = (int) $userId;
		$db     = $this->db;
		$query  = $db->getQuery(true);

		$query->select('COUNT(*)')
		      ->from($db->quoteName('#__privacy_consents'))
		      ->where($db->quoteName('user_id') . ' = :userid')
		      ->where($db->quoteName('subject') . ' = ' . $db->quote('PLG_SYSTEM_PRIVACYCONSENT_SUBJECT'))
		      ->where($db->quoteName('state') . ' = 1')
		      ->bind(':userid', $userId, \Joomla\Database\ParameterType::INTEGER);
		$db->setQuery($query);

		return (int) $db->loadResult() > 0;
	}
}