<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Actionlog\Loginguard\Extension;

defined('_JEXEC') || die();

use Akeeba\Component\LoginGuard\Administrator\Event\NotifyActionLog;
use Akeeba\Component\LoginGuard\Administrator\View\Methods\HtmlView;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Factory;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Component\Actionlogs\Administrator\Plugin\ActionLogPlugin as BaseActionLogPlugin;
use Joomla\Event\SubscriberInterface;

/**
 * LoginGuard integration with Joomla's User Actions Log
 *
 * @since  3.1.2
 */
class Loginguard extends BaseActionLogPlugin implements SubscriberInterface
{
	/**
	 * The Joomla application object.
	 *
	 * @var   CMSApplication
	 * @since 5.0.0
	 */
	protected $app;

	/**
	 * Autoload the language strings for this plugin
	 *
	 * @var    boolean
	 * @since  6.0.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * The name of the component we are logging actions for
	 *
	 * @var   string
	 * @since 5.0.0
	 */
	private $defaultExtension = 'com_loginguard';

	/** @inheritdoc */
	public static function getSubscribedEvents(): array
	{
		return [
			'onComLoginGuardViewMethodsAfterDisplay'                  => 'onComLoginGuardViewMethodsAfterDisplay',
			'onComLoginguardCaptiveShowCaptive'                       => 'onComLoginguardCaptiveShowCaptive',
			'onComLoginguardCaptiveShowSelect'                        => 'onComLoginguardCaptiveShowSelect',
			'onComLoginguardCaptiveValidateFailed'                    => 'onComLoginguardCaptiveValidateFailed',
			'onComLoginguardCaptiveValidateInvalidMethod'             => 'onComLoginguardCaptiveValidateInvalidMethod',
			'onComLoginguardCaptiveValidateSuccess'                   => 'onComLoginguardCaptiveValidateSuccess',
			'onComLoginguardControllerMethodAfterRegenbackupcodes'    => 'onComLoginguardControllerMethodAfterRegenbackupcodes',
			'onComLoginguardControllerMethodBeforeAdd'                => 'onComLoginguardControllerMethodBeforeAdd',
			'onComLoginguardControllerMethodBeforeDelete'             => 'onComLoginguardControllerMethodBeforeDelete',
			'onComLoginguardControllerMethodBeforeEdit'               => 'onComLoginguardControllerMethodBeforeEdit',
			'onComLoginguardControllerMethodBeforeSave'               => 'onComLoginguardControllerMethodBeforeSave',
			'onComLoginguardControllerMethodsBeforeDisable'           => 'onComLoginguardControllerMethodsBeforeDisable',
			'onComLoginguardControllerMethodsBeforeDontshowthisagain' => 'onComLoginguardControllerMethodsBeforeDontshowthisagain',
			'onComLoginguardControllerConvertAfterConvert'            => 'onComLoginguardControllerConvertAfterConvert',
		];
	}

	/**
	 * Logs converting from Joomla's MFA
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerConvertAfterConvert(NotifyActionLog $event): void
	{
		$this->logUserAction('', 'PLG_ACTIONLOG_LOGINGUARD_ACTION_CONVERT', 'com_loginguard');
	}

	/**
	 * Logs showing the Multi-factor Authentication selection Method
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardCaptiveShowSelect(NotifyActionLog $event): void
	{
		$this->logUserAction('', 'PLG_ACTIONLOG_LOGINGUARD_ACTION_CAPTIVE_SELECT', 'com_loginguard');
	}

	/**
	 * Logs showing the Captive login page
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardCaptiveShowCaptive(NotifyActionLog $event): void
	{
		/**
		 * @var   string $methodTitleEscaped
		 */
		[$methodTitleEscaped] = $event->getArguments();

		$this->logUserAction($methodTitleEscaped, 'PLG_ACTIONLOG_LOGINGUARD_ACTION_CAPTIVE_CAPTIVE', 'com_loginguard');
	}

	/**
	 * Log displaying a user's Multi-factor Authentication Methods
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginGuardViewMethodsAfterDisplay(NotifyActionLog $event): void
	{
		/**
		 * @var   HtmlView $view
		 */
		[$view] = $event->getArguments();
		$layout = $view->getLayout();
		$key    = 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHODS_SHOW';

		if ($layout == 'firsttime')
		{
			$key = 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHODS_FIRSTTIME';
		}

		$this->logUserAction('', $key, 'com_loginguard');
	}

	/**
	 * Log regenerating backup codes
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodAfterRegenbackupcodes(NotifyActionLog $event): void
	{
		$this->logUserAction('', 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHOD_REGENBACKUPCODES', 'com_loginguard');
	}

	/**
	 * Log adding a new Multi-factor Authentication Method
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodBeforeAdd(NotifyActionLog $event): void
	{
		/**
		 * @var User   $user
		 * @var string $method
		 */
		[$user, $method] = $event->getArguments();

		$this->logUserAction([
			'method'    => $method,
			'user_id'   => $user->id,
			'otheruser' => $user->username,
		], 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHOD_ADD', 'com_loginguard');
	}

	/**
	 * Log editing a Multi-factor Authentication Method
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodBeforeEdit(NotifyActionLog $event): void
	{
		/**
		 * @var int  $id
		 * @var User $user
		 */
		[$id, $user] = $event->getArguments();

		$this->logUserAction([
			'id'        => $id,
			'user_id'   => $user->id,
			'otheruser' => $user->username,
		], 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHOD_EDIT', 'com_loginguard');
	}

	/**
	 * Log removing a Multi-factor Authentication Method
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodBeforeDelete(NotifyActionLog $event): void
	{
		/**
		 * @var int  $id
		 * @var User $user
		 */
		[$id, $user] = $event->getArguments();

		$this->logUserAction([
			'id'        => $id,
			'user_id'   => $user->id,
			'otheruser' => $user->username,
		], 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHOD_DELETE', 'com_loginguard');
	}

	/**
	 * Log saving a Multi-factor Authentication Method
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodBeforeSave(NotifyActionLog $event): void
	{
		/**
		 * @var int  $id
		 * @var User $user
		 */
		[$id, $user] = $event->getArguments();

		$this->logUserAction([
			'id'        => $id,
			'user_id'   => $user->id,
			'otheruser' => $user->username,
		], 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHOD_SAVE', 'com_loginguard');
	}

	/**
	 * Log completely disabling Multi-factor Authentication
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodsBeforeDisable(NotifyActionLog $event): void
	{
		/**
		 * @var User $user
		 */
		[$user] = $event->getArguments();

		$this->logUserAction([
			'user_id'   => $user->id,
			'otheruser' => $user->username,
		], 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHODS_DISABLE', 'com_loginguard');
	}

	/**
	 * Log opting out of Multi-factor Authentication
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardControllerMethodsBeforeDontshowthisagain(NotifyActionLog $event): void
	{
		/**
		 * @var User $user
		 */
		[$user] = $event->getArguments();

		$this->logUserAction([
			'user_id'   => $user->id,
			'otheruser' => $user->username,
		], 'PLG_ACTIONLOG_LOGINGUARD_ACTION_METHODS_DONTSHOWTHISAGAIN', 'com_loginguard');
	}

	/**
	 * Log Multi-factor Authentication failure due to invalid Method
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardCaptiveValidateInvalidMethod(NotifyActionLog $event): void
	{
		$this->logUserAction('', 'PLG_ACTIONLOG_LOGINGUARD_ACTION_VALIDATE_INVALID_METHOD', 'com_loginguard');
	}

	/**
	 * Log Multi-factor Authentication failure
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardCaptiveValidateFailed(NotifyActionLog $event): void
	{
		/**
		 * @var   string $methodTitle
		 */
		[$methodTitle] = $event->getArguments();

		$this->logUserAction(htmlspecialchars($methodTitle), 'PLG_ACTIONLOG_LOGINGUARD_ACTION_VALIDATE_FAILED', 'com_loginguard');
	}

	/**
	 * Log Multi-factor Authentication success
	 *
	 * @param   NotifyActionLog  $event
	 *
	 * @return  void
	 * @since   6.0.0
	 */
	public function onComLoginguardCaptiveValidateSuccess(NotifyActionLog $event): void
	{
		/**
		 * @var   string $methodTitle
		 */
		[$methodTitle] = $event->getArguments();

		$this->logUserAction(htmlspecialchars($methodTitle), 'PLG_ACTIONLOG_LOGINGUARD_ACTION_VALIDATE_SUCCESS', 'com_loginguard');
	}

	/**
	 * Log a user action.
	 *
	 * This is a simple wrapper around self::addLog
	 *
	 * @param   string|array  $title               Language key for title or an array of additional data to record in
	 *                                             the audit log.
	 * @param   string        $logText             Language key describing the user action taken.
	 * @param   string|null   $extension           The name of the extension being logged (default: use
	 *                                             $this->defaultExtension).
	 * @param   User|null     $user                User object taking this action (default: currently logged in user).
	 *
	 * @return  void
	 *
	 * @see     self::addLog
	 * @since   5.0.0
	 */
	private function logUserAction($title, string $logText, ?string $extension = null, ?User $user = null): void
	{
		// Make sure I am in the front- or backend application
		if (empty($this->app) || (!$this->app->isClient('administrator') && !$this->app->isClient('site')))
		{
			return;
		}

		// Get the user if not defined
		$user = $user ?? (
			$this->app->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0)
			);


		// No log for guests
		if (empty($user) || ($user->guest))
		{
			return;
		}

		// Default extension if none defined
		$extension = $extension ?? $this->defaultExtension;

		$message = [
			'username'    => $user->username,
			'accountlink' => 'index.php?option=com_users&task=user.edit&id=' . $user->id,
		];

		if (!is_array($title))
		{
			$title = [
				'title' => $title,
			];
		}

		$message = array_merge($message, $title);

		$this->addLog([$message], $logText, $extension, $user->id);
	}
}
